#include "seg_global.h"
#include "seg.h"


/******************************************************************************/
/* Save and read Tree_Info*/ 
/******************************************************************************/
static void SaveTreeInfo( FILE *fd, struct Tree_Info *info )
{
  fprintf(fd, "nclassess=%d\n", info->nclasses) ;
  fprintf(fd, "dimen=%d\n", info->dimen) ;
  fprintf(fd, "nTree=%d\n", info->nTree) ;
  fprintf(fd, "RangeF=%d\n", info->RangeF) ;
}


static void ReadTreeInfo( FILE *fd, struct Tree_Info *info )
{
  fscanf(fd, "nclassess=%d\n", &(info->nclasses)) ;
  if(info->nclasses>MAX_IMAGE_CLASS) {
      fprintf(stderr, "ERROR:  number of classes must be <= %d\n",
          MAX_IMAGE_CLASS);
      exit(-1);
  }

  fscanf(fd, "dimen=%d\n", &(info->dimen)) ;
  info->Margin = (int) ( sqrt((double) info->dimen) ) ;
  if(info->Margin * info->Margin != info->dimen) {
      fprintf(stderr, "ERROR: CPT dimen %d must be the square of an integer\n",
          info->dimen) ;
      exit(-1) ;
  } else {
      info->Margin = (info->Margin -1)/2 ;
  }
      
  fscanf(fd, "nTree=%d\n", &(info->nTree)) ;
  fscanf(fd, "RangeF=%d\n", &(info->RangeF)) ;
  
  if(vlevel>0) {
      printf("\tCPT info\n") ;
      printf("\t\tnclasses = %d\n", info->nclasses) ;
      printf("\t\tdiman = %d\n", info->dimen) ;
      printf("\t\tmargin = %d\n", info->Margin) ;
      printf("\t\tnTree = %d\n", info->nTree) ;
      printf("\t\trangeF = %d\n", info->RangeF) ;
  }
}


/******************************************************************************/
/* Save this tree, in the order of parent, left, right.*/  
/* If missing, save "***" */
/******************************************************************************/
static void SaveDecTreePar(FILE *fd, struct TreeNode *x, int nClasses, 
    int nNeighb)
{
  int i,j ;
  double **Hist ;
  double **T ;
    
  if (x!= NULL)  {        /* recursively save tree nodes in DESENDING ORDER */
  
    Hist = x->Hist ;
    T = x->T ;
        
    fprintf(fd,"======  %d \n", x->nodeNumber) ;
        
    fprintf(fd,"( <0)") ;
    for(j=0; j<nClasses; j++) fprintf(fd," %d=%lf", j,Hist[0][j]);
    fprintf(fd,"\n") ;
                    
    fprintf(fd,"(>=0)") ;
    for(j=0; j<nClasses; j++) fprintf(fd," %d=%lf", j,Hist[1][j]);
    fprintf(fd,"\n") ;
          
    fprintf(fd, "T0=%lf\n", x->T0) ;
    for(i=0; i<nNeighb; i++) {
        for(j=0; j<(nClasses-1); j++) fprintf(fd,"%lf  ",T[i][j]) ;
        fprintf(fd,"\n") ;
    }
          
    SaveDecTreePar(fd, x->left, nClasses, nNeighb);
    SaveDecTreePar(fd, x->right, nClasses, nNeighb);
  } else {
      fprintf(fd, "***\n") ;
  }
}




/******************************************************************************/
/* construct a Decesion Tree node */
/******************************************************************************/
struct TreeNode *GenDecTreeNode(struct TreeNode *parentPt, char LR,
  int nClasses, int nNeighb, int rangeF)
  /* char LR -- generate left "L" child or right "R" child */
{
    struct TreeNode *newNode ;
                    
    if(vlevel >100) printf("nCl=%d, nNb=%d, Rf=%d\n", nClasses,nNeighb,rangeF);
    
    newNode=(struct TreeNode *) get_spc(1,sizeof(struct TreeNode)) ;
         newNode->left=NULL ;
    newNode->right=NULL ;
    newNode->Hist = (double **) get_mtrx(rangeF, nClasses, sizeof(double)) ;
    newNode->T = (double **) get_mtrx(nNeighb,nClasses-1,sizeof(double)) ; 

    if(parentPt==NULL) {    /* The tree is empty */
        newNode->parent=NULL ;
        newNode->level = 0 ;

    } else { /* the tree is not empty */
        newNode->parent=parentPt ;
        newNode->level = (parentPt->level) +1 ;
        
        if(((parentPt->left)==NULL) && (LR=='L')) parentPt->left=newNode ;
        else if(((parentPt->right)==NULL) && (LR=='R')) parentPt->right=newNode;
        else {
            fprintf(stderr, "ERROR: Parent has the child, %c.\n",LR) ;
            exit(-1) ;
        }
    }
    return(newNode) ;
}






/******************************************************************************/
/* read saved decision tree parameters */
/* if depth is 0, Read... return a pt to the root of the decision tree */
/******************************************************************************/

static struct TreeNode* ReadTreeNodePar(
  FILE *fd,
  struct TreeNode *parentPt,    /* DecTreeNode for segmentation part */
  int depth, 
  char LR,    /* generate left "L" child or right "R" child */
  int nClasses,
  int nNeighb,
  int rangeF
)
{
  int i,j,k, tmp;
  double **Hist ;
  char Str[80],Str2[80], Char ;
  struct TreeNode *Node ;
  double **TT ;
    
  if(fgets(Str,80,fd) ==NULL) {
      fprintf(stderr, "fgets error! ReadTreePar\n") ;
      exit(-1) ;
  }
      
  sscanf(Str,"%s",Str2) ;
  if(vlevel>100) printf("level=%d, %s, %d, %s, %d\n",
      depth, Str,strlen(Str), Str2,strlen(Str2)) ;

  if((strcmp(Str2,"***")!=0)&&(!feof(fd)))  {        /* not an empty node */
      Node = GenDecTreeNode(parentPt, LR, nClasses, nNeighb, rangeF) ;
      
    if((Node->level) != depth) {
        fprintf(stderr, "Level masses up! ReadTreeStrHist") ;
        fprintf(stderr, " L%d, D%d\n", Node->level, depth) ;
        exit(-1) ;
    }
        
    /* read in histogram */
    Hist = Node->Hist ;
    for(k=0; k<rangeF; k++) {
        fscanf(fd, "(%c%c%c)", &Char,&Char,&Char) ;
        for(j=0; j<nClasses; j++) {
             fscanf(fd," %d=%lf", &tmp,&(Hist[k][j]));
             if(tmp!=j) {
                 fprintf(stderr, "ERROR: wrong format %d != %d\n", tmp, j) ;
                 fprintf(stderr, "ERROR: ReadTreeNodePar()\n") ;
                 exit(-1) ;
             }
        }
        fscanf(fd,"\n") ;
    }
    
        
    /* read linear partition function */
    TT=Node->T ;
    fscanf(fd,"T0=%lf",&(Node->T0)) ;
    for(i=0; i<nNeighb; i++) 
    for(j=0; j<(nClasses-1); j++) fscanf(fd,"%lf",&(TT[i][j])) ;
    fscanf(fd,"\n") ;
    
    /* test */
    if(vlevel >100) PrintNode(stdout, Node, nClasses, nNeighb, rangeF) ;
        
    
    /* recursively read left child, right child */
    ReadTreeNodePar(fd, Node, depth+1, 'L', nClasses, nNeighb, rangeF) ;
    ReadTreeNodePar(fd, Node, depth+1, 'R', nClasses, nNeighb, rangeF) ;

    if(depth != 0) {    /* not the root */
        return(NULL) ;
    } else {    /* the root */
        return(Node) ;
    }
  } else {
    if(vlevel>100) printf("IS ***\n") ;  
    return(NULL) ;
  }
}



/******************************************************************************/
/******************************************************************************/
void PrintNode( FILE *fd, struct TreeNode *x, int nClasses, int nNeighb, 
    int rangeF )
{

  int i,j ;
  double **Hist ;
  double **T ;
  
  if(rangeF != 2) {
      fprintf(stderr, "ERROR:  PrintNode can only print node of a binary\n") ;
      fprintf(stderr, "ERROR:  tree, not %d-ary tree\n", rangeF) ;
      exit(-1) ;
  }
    
  if (x!= NULL)  {        /* recursively save tree nodes in DESENDING ORDER */
  
    Hist = x->Hist ;
    T = x->T ;
        
    fprintf(fd,"======\n") ;
        
    fprintf(fd,"( <0)") ;
    for(j=0; j<nClasses; j++) fprintf(fd," %d=%lf", j,Hist[0][j]);
    fprintf(fd,"\n") ;
                    
    fprintf(fd,"(>=0)") ;
    for(j=0; j<nClasses; j++) fprintf(fd," %d=%lf", j,Hist[1][j]);
    fprintf(fd,"\n") ;
          
      fprintf(fd, "T0=%lf\n", x->T0) ;
      for(i=0; i<nNeighb; i++) {
          for(j=0; j<(nClasses-1); j++) fprintf(fd,"%lf  ",T[i][j]) ;
          fprintf(fd,"\n") ;
      }
          
  } else {
      fprintf(fd, "***\n") ;
  }

}


/******************************************************************************/
/******************************************************************************/
void ShowDecTree( FILE *fd, struct TreeNode *x, int depth, int nClasses,
    int rangeF )
{
  int i,j,k;
    
  if (x!= NULL)  {        /* recursively print tree in DESENDING ORDER */
  
    ShowDecTree(fd, x->right,depth+1, nClasses, rangeF);

     /* PRINT BLANK SPACES, so that the printed order*/
    /*creates a virual tree shape, 90 degrees rotated with parent as a pivot */
    
    for(i=0; i<rangeF; i++) {
        for(k=0;k<depth*4;k++) fprintf(fd, " ");
        for(j=0; j<nClasses; j++) fprintf(fd," %d=%lf",j,((x->Hist)[i][j]));
        fprintf(fd,"\n") ;
    }
    
    ShowDecTree(fd, x->left,depth+1, nClasses, rangeF);
  }
}


/******************************************************************************/
/* Save and read CP_Tree parameters */
/******************************************************************************/
void SaveCPT_Par( struct CP_Trees *CPT, char *filename )
{
  FILE *fd ;
  char fn[128] ;
  int i,j, nTree, nClasses, nNeighb;
  struct Trees *treePt ;
  
  if(filename != NULL) {
	  sprintf(fn,"%s.cpt", filename) ;
	  if(vlevel>3) printf("\tsaving %s ......\n",fn) ;

	  fd = fopen(fn,"w") ;
	  if(fd==NULL) FileOpenErr(fn) ;
	  
  } else fd = stdout ;
  
  fprintf(fd, "nlevels=%d\n", CPT->nlevels) ;
  SaveTreeInfo( fd, &(CPT->info) ) ;
  
  nTree = (CPT->info).nTree ;
  nNeighb = (CPT->info).dimen ;
  nClasses = (CPT->info).nclasses ;
  
  for(i=1; i<(CPT->nlevels); i++) {
      fprintf(fd, "\nlevel=%d\n", i) ;
      
      treePt = &((CPT->trees)[i]) ;
      for(j=0; j<nTree; j++) {
          fprintf(fd, "\n+++++ Tree %d +++++\n", j) ;
          SaveDecTreePar(fd, (treePt->roots)[j], nClasses, nNeighb) ;
      }
  }
  
  if(filename != NULL) fclose(fd) ;
}


/******************************************************************************/
/* read saved decision tree parameters */
/******************************************************************************/

void ReadCPT_Par(struct CP_Trees *CPT, char *filename)
{
  FILE *fd ;
  char fn[128] ;
  int i,j,tmp, nTree, nClasses, nNeighb, rangeF;
  struct Trees *treePt ;
  
  sprintf(fn,"%s.cpt", filename) ;
  if(vlevel>3) printf("\treading and initializing %s ......\n",fn) ;

  fd = fopen(fn,"r") ;
  if(fd==NULL) FileOpenErr(fn) ;
  
  fscanf(fd, "nlevels=%d\n", &(CPT->nlevels)) ;
  ReadTreeInfo( fd, &(CPT->info) ) ;
  
  nTree = (CPT->info).nTree ;
  nNeighb = (CPT->info).dimen ;
  nClasses = (CPT->info).nclasses ;
  rangeF = (CPT->info).RangeF ;
  CPT->trees = (struct Trees*) get_spc(CPT->nlevels, sizeof(struct Trees)) ;
  
  for(i=1; i<(CPT->nlevels); i++) {
      
      fscanf(fd, "\nlevel=%d\n", &tmp) ;
      if(i!=tmp) {
          fprintf(stderr, "ERROR:  Wrong format 1, %d != %d\n", i, tmp) ;
          fprintf(stderr, "ERROR:  ReadCPT_Par()\n") ;
          exit(-10) ;
      }
      
      treePt = &((CPT->trees)[i]) ;
      treePt->nLeaves = NULL ;
      treePt->roots = (struct TreeNode **) get_spc(nTree, 
          sizeof(struct TreeNode *)) ;
      treePt->infoPt = &(CPT->info) ;
      
      for(j=0; j<nTree; j++) {
          fscanf(fd, "\n+++++ Tree %d +++++\n", &tmp) ;
          if(j!=tmp) {
              fprintf(stderr, "ERROR:  Wrong format 2, %d != %d\n", j, tmp) ;
              fprintf(stderr, "ERROR:  ReadCPT_Par()\n") ;
              exit(-10) ;
          }
      
          (treePt->roots)[j] = ReadTreeNodePar(fd, (struct TreeNode *) NULL,
              0, 'L', nClasses, nNeighb, rangeF) ;  
                
          if(vlevel>50) {
              printf("\n\nDecTree Read:  level=%d, no.=%d  \n", i,j);
              ShowDecTree(stdout, (treePt->roots)[j], 0, nClasses, rangeF);
          }

      }
  }
  
  fclose(fd) ;
}


/******************************************************************************/
/* free a node  */                 
/******************************************************************************/
static void TreeFreeNode(struct TreeNode* node)
{		
	if(node->Hist != NULL) free_mtrx((void**) node->Hist) ;
	if(node->T != NULL) free_mtrx((void**) node->T) ;
  	if(node != NULL) free((void*) node) ;
}


	
	
/******************************************************************************/
/* free node below  */                 
/******************************************************************************/
static void FreeNodeBelow(struct TreeNode* node)
{
	struct TreeNode *child ;
				
    child = node->left ;
    if(child != NULL) FreeNodeBelow(child) ;
     	
    child = node->right ;
    if(child != NULL) FreeNodeBelow(child) ;
     	
	TreeFreeNode(node) ;     	
}

/******************************************************************************/
/* read saved decision tree parameters */
/******************************************************************************/

void FreeCPT_Par(struct CP_Trees *CPT)
{
  int i,j ;
  struct Trees *treePt ;
    
  for(i=1; i<(CPT->nlevels); i++) {
      treePt = &((CPT->trees)[i]) ;
      if(treePt->roots != NULL) {
          for(j=0; j<(CPT->info).nTree; j++) FreeNodeBelow((treePt->roots)[j]) ;
          free( (void *) treePt->roots) ;
          free( (void *) treePt->nLeaves) ;
      }
      if(treePt->nLeaves != NULL) free( (void *) treePt->nLeaves) ;
  }
  
  free( (void *) CPT->trees ) ;
}

