#ifndef SEG_H
#define SEG_H

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>
#include "../../utilities/src/util.h"
#include "../../training/src/TreeStruct.h"
#include "sig.h"


#define LIKELIHOOD double
#define MAX_IMAGE_CLASS 64    /* the # of files can be opened simutanously */
                              /* refer to how to generate samples for GMM */
#define PI 3.141592654
#define FALSE 0
#define TRUE 1

/* using a look up table to substitute for exp() */
#define EXP_MAX   0.0        /* since dealing with prob, MAX sets to 0 */
#define EXP_MIN -20.0        /* EXP_MIN=-10 does not work for up() */
#define EXPnBins 20000       /* When EXPnBins=20000,  */
                             /* exp_lookup() = exp() * (1 +/- 0.001) */

/* using a look up table to substitute for exp() */
#define MIN_LOG  0.000001    /* if x < MIN_LOG,  log(x) <== log(MIN_LOG) */
#define LOG_MAX  1.0        /* since dealing with prob, MAX sets to 1.0 */
#define LOGEnBins 50000      /* When LOGEnBins=50000,  */
                             /* log(0.001) - log(0.00098) = 0.0202*/



/* */
/* #ifdef CLASS_TYPE_SHORT */
/* # define IMAGE_CLASS unsigned short */
/* # define MAX_IMAGE_CLASS 65535 */
/* #else */
/* # define IMAGE_CLASS unsigned char */
/* # define MAX_IMAGE_CLASS 255 */
/* #endif */
/* */


/******************************************************************************/
/* Quadtree parameters */
/******************************************************************************/
#ifndef QTR_DEF
#define QTR_DEF

struct Par_QTR
{
   int nclasses ;
   int levels ;           /* number of resolutions */
   double ***Alpha0 ;     /* trasition para between pixel and
                             its father [level][fine_class][coarse_class] */
} ;  
#endif  /* QTR_DEF */


/******************************************************************************/
/* GMM model parameters */
/******************************************************************************/
#ifndef  GMM_DEF
#define  GMM_DEF

struct Par_GMM
{
    char feature_name[128] ;
    int  dimen ;
    int  nclasses ;
    int  MaxLevels ;	         /* Max level of features needed, or allowed */
    int  Rt ;	                 /* subsampling rate when extracting features */

    int  *existSig ;	         /* =1, if a signature exists at that level */ 
    struct SigSet *_Sig ; 	 /* signature at a level */
  
    void (*GenerateFunc)() ;	 /* Generating Function */
} ;
#endif  /* GMM_DEF */



/******************************************************************************/
/* General parameters */
/******************************************************************************/
#ifndef  GEN_DEF
#define  GEN_DEF

struct ModelPar
{
    struct Par_QTR QTr_par ;
    struct Par_GMM GMM_par ;
    struct CP_Trees CPT ;
    
    double *DataWeight ;    /* DataWeight[level] range from [0,1].   
                               LL +  DataWeight * P(X|parent)  */
	
} ;
#endif  /* GEN_DEF */




/* BigPyramid
*/
struct BigPyramid
{
	/* basic parameters */
	int level ;	/* number of levels of the BigPyramid */
	int *nRow ;	/* number of rows of each level */
	int *nCol ;	/* number of columns of each level */

	/* segmentation related parameters */
	double ****LProb ; 	/* log probability of Xn given Yn
				           _LProb[level][j][k][class] */	
	double ****LLikeH ; 	/* LLikeH[level][j][k][class], log likelihood*/

	unsigned char ***_Segment ; 	/* _Segment[d][i][j], segment at 
					                   level d, row i, column j */

} ;
	


/* MAP.c */

void MAP_Pyramid(struct BigPyramid *Pyramid, struct ModelPar *Par, char *outfn,
    int X0, int Y0); 
void SegPyramidModel(
  struct BigPyramid *Pyramid,
  struct ModelPar *Par, 
  int K,  /*  K the current level */
  FILE *outfd
);
int SegMaxProb(
  struct ModelPar *Par,
  double *ll,
  double *TranProb,
  FILE *outfd,
  int K    /* level */
);
void SegCoarsestResolution(struct BigPyramid* Pyramid, int nClasses) ;
double LogCalculate( double p, char *Str) ;
void MAP_QuadTree(struct BigPyramid *Pyramid, struct ModelPar *Par,
     char *outfn, int X0, int Y0) ;


/* ML */

void ML_Test(struct BigPyramid* Pyramid, struct ModelPar *Par, char *outfn) ;


/* alloc.c */

char *G_malloc(int n) ;
char *G_calloc(int n, int m);
char *G_realloc(char *b, int n);
double *G_alloc_vector(int n);
double **G_alloc_matrix(int rows, int cols);
void G_free_vector(double *v);
void G_free_matrix(double **m);
int *G_alloc_ivector(int n);
int **G_alloc_imatrix(int rows, int cols);
void G_free_ivector(int *v);
void G_free_imatrix(int **m);


/* cntxProb.c */

void CoarseNeighbors(
   struct CP_Trees *CPT,
   unsigned char **SegFather,
   int nrowF, int ncolF,
   int ii,  int jj,      /* current pixel indices */
   int *Neighbors
) ;

double* ContexProb(
  int *X,    /* known classification of coarse neighbors */
  struct TreeNode *DecTrNode, /* DecTreeNode -- decision tree structure */
  int nClasses,
  int nNeighb
) ;


/* cntx_io.c */

int write_segmentation(unsigned char **seg, int height, int width, char *fn) ;
void SaveCPT_Par(struct CP_Trees *CPT, char *filename) ;
void ReadCPT_Par(struct CP_Trees *CPT, char *filename) ;
void FreeCPT_Par(struct CP_Trees *CPT) ;
void PrintNode( FILE *fd, struct TreeNode *x, int nClasses, int nNeighb, 
    int rangeF ) ;
struct TreeNode *GenDecTreeNode(struct TreeNode *parentPt, char LR,
  int nClasses, int nNeighb, int rangeF) ;
void ShowDecTree( FILE *fd, struct TreeNode *x, int depth, int nClasses,
    int rangeF ) ;



/* decimate.c */

void upPropagate(struct BigPyramid* Pyramid, struct ModelPar *Par) ;


/* eigen.c */

void eigenNoEigenVector(double **M, double *lambda, int n)  ;
/* No eigen vectors are returned */
/* Computes eigenvalues (and eigen vectors if desired) for	*/
/*  symmetric matices. 						*/
/* double 	**M;*/	 /* Input matrix */
/* double  *lambda; */ /* Output eigenvalues */
/* int	n;	     */    /* Input matrix dimension */


/* feature.c */

void sub_sample(double **InImg, int wd, int ht, int sRate,
  double ***OutImg, int *wdPt, int *htPt ) ;
void PrintFeatureCh(struct Par_GMM *GMM_par) ;
void FeatureInitial(
  eisl_img *img_in, 
  struct Par_GMM *GMM_par,
  int *ncol, int *nrow
) ;
void FindNeighbors(int **neighbors, int i0, int j0, int ht, int wd, int Dimen);



/* featureProb.c */

void FeatureProb(
  struct BigPyramid* Pyramid,
  eisl_img* imgPt, 	/* input images */
  struct Par_GMM *GMM_par
) ;
void FeatProbOutput(struct BigPyramid* Pyramid, struct Par_GMM *GMM_par,
    char *fn) ;
void FeatProbRead(struct BigPyramid* Pyramid, struct Par_GMM *GMM_par,
    char *fn) ;


/* invert.c */

int invert(double **a, int n) ;


/* model.c */

int extract_init(struct SigSet *S) ;
void extract(
    double ***img,    /* multispectral image, img[band][i][j] */
    int nrow, int ncol,
    double ***ll,            /* log likelihood, ll[i][j][class] */
    struct SigSet *S        /* class signatures */
) ;




/* seg_io.c */

void ReadQuadTreePar( struct Par_QTR *QTr_par, char *alpha_fn ) ;
void WriteQuadTreePar( FILE *fd, struct Par_QTR *QTr_par) ;
void FreeQuadTreePar( struct Par_QTR *QTr_par ) ;
void ReadGMM_Par( struct Par_GMM *GMM_par, char *GMM_fn ) ;
void WriteGMM_Par( struct Par_GMM *GMM_par, char *GMM_fn ) ;
void FreeGMM_Par( struct Par_GMM *GMM_par ) ;
void OutputSeg( struct BigPyramid *Pyramid,char *fn, char *String ) ;
void ReadDataWeight( struct ModelPar *Par, char *fn ) ;
void WriteDataWeight( struct ModelPar *Par,  char *fn ) ;


/* strip.c */

void G_strip (register char *buf) ;



/* u_pyramid.c */

void ConstructPyramid (int nrow, int ncol, struct BigPyramid *bigPyramid, 
    struct ModelPar *Par) ;
void FreePyramid (struct BigPyramid *bigPyramid) ;
int levels(int nrow, int ncol) ;
double exp_lookup( double x ) ;
double logE_lookup( double x ) ;




#endif   /* SEG_H */
