#include "seg.h"
#include "seg_global.h"

/**************************************************************************/
/*    Read and initialization of quadtree parameters*/
/**************************************************************************/
void ReadQuadTreePar( struct Par_QTR *QTr_par, char *alpha_fn ) 
{
  FILE *fd ;
  char fn[128] ;
  int m1,m2, i, index;
  int nClasses ;
  double tmp ;
  double ***alpha ;

  sprintf(fn,"%s.qtr",alpha_fn) ;
  if(vlevel>3) printf("\treading and initializing %s ......\n",fn) ;

  fd = fopen(fn,"r") ;
  if(fd==NULL) FileOpenErr(fn) ;
  
  fscanf(fd, "nClasses=%d\n", &(QTr_par->nclasses)) ;
  fscanf(fd, "levels=%d\n", &(QTr_par->levels)) ;

  if(QTr_par->nclasses>MAX_IMAGE_CLASS) {
      fprintf(stderr, "ERROR:  number of classes must be <= %d\n",
          MAX_IMAGE_CLASS);
      exit(-1);
  }
      
  (QTr_par->Alpha0)=(double ***) mget_spc(QTr_par->levels, sizeof(double*)) ;
  alpha = QTr_par->Alpha0 ;
  nClasses = QTr_par->nclasses ;
    
  for(i=1; i<(QTr_par->levels); i++) {

      if(feof(fd)) FeofErr("ReadAlpha()") ;
      fscanf(fd,"level=%d\n", &index) ;

      if(index != i) {
          fprintf(stderr,"ERROR: Wrong format for quadtree par i%d, %d.\n",
              i, index) ;
          exit(-10) ;
      }
      
      alpha[i] = (double**) get_img(nClasses,nClasses,sizeof(double)) ;
      
      for(m2=0; m2<nClasses; m2++)  /* classes at coarse resolution */
      for(m1=0; m1<nClasses; m1++) { /* classes at fine resolution */

          if(feof(fd)) FeofErr("ReadAlpha()") ;
          fscanf(fd, "%lf", &tmp) ;

/***********  when log(alpha) is saved *************/          
          alpha[i][m1][m2] = exp(tmp) ; 
         
/***********  when  alpha is saved ****************          

          if(tmp != 0.0)
              alpha[i][m1][m2] = tmp ; 
          else {
              alpha[i][m1][m2] = MIN_LOG ; 
          }
**************************************************/
     
     }
     fscanf(fd, "\n") ;
     
  }
  fclose(fd) ;

  if(vlevel>100) {
      printf("Quadtree parameters:\n") ;
      WriteQuadTreePar(stdout, QTr_par) ;

  } else if(vlevel>0) {
      printf("\tQuadtree parameters:\n") ;
      printf("\t\tnlevels = %d\n", QTr_par->levels) ;
      printf("\t\tnclasses = %d\n", QTr_par->nclasses) ;
  }
}



/*************************************************************************/
/*    write out QTr_par*/
/*************************************************************************/
void WriteQuadTreePar(FILE *fd, struct Par_QTR *QTr_par) 
{
  int m1,m2, i;
  double ***alpha ;

  if(vlevel>3) printf("\twriting alpha ......\n") ;
  
  alpha = QTr_par->Alpha0 ;
  
  fprintf(fd, "nClasses=%d\n", QTr_par->nclasses) ;
  fprintf(fd, "levles=%d\n", QTr_par->levels ) ;
  
  for(i=1; i<(QTr_par->levels); i++) {
      fprintf(fd, "level=%d\n", i) ;
      for(m2=0; m2<(QTr_par->nclasses); m2++) { /* at coarse resolution */
        for(m1=0; m1<(QTr_par->nclasses); m1++) { /* at fine resolution */
              fprintf(fd,"   %lf", alpha[i][m1][m2]) ;
        }
       fprintf(fd,"\n") ;
      }
  }
}



/*****************************************************************************/
/*    free quadtree parameters */
/*****************************************************************************/
void FreeQuadTreePar( struct Par_QTR *QTr_par ) 
{
  int i ;
  double ***alpha ;
  
  alpha = QTr_par->Alpha0 ;
  
  for(i=1; i<(QTr_par->levels); i++) {
      if( alpha[i] != NULL ) free_img( (void **) alpha[i] ) ;
  }
  
  free( (void *) alpha) ;
}





/**************************************************************************/
/*    Read and initialization of Gaussian mixture model parameters*/
/**************************************************************************/
void ReadGMM_Par( struct Par_GMM *GMM_par, char *GMM_fn ) 
{
  FILE *fd, *parfd ;
  char fn[128], parfn[80] ;
  struct SigSet *S_Pt ;
  int i, MaxFL ;

  sprintf(fn,"%s.gmm", GMM_fn) ;
  if(vlevel>3) printf("\treading and initializing %s ......\n",fn) ;
  fd = fopen(fn,"r") ;
  if(fd==NULL) FileOpenErr(fn) ;
  
  fscanf(fd, "NAME %s\n",  GMM_par->feature_name) ;
  fscanf(fd, "nbands=%d\n", &(GMM_par->dimen)) ;
  fscanf(fd, "nClasses=%d\n", &(GMM_par->nclasses)) ;
  fscanf(fd, "MaxLevels=%d\n", &(GMM_par->MaxLevels)) ;
  
  GMM_par->existSig = (int *) mget_spc(GMM_par->MaxLevels, sizeof(int)) ;
  GMM_par->_Sig = (struct SigSet*) mget_spc(GMM_par->MaxLevels,
      sizeof(struct SigSet)) ;
      
    
  /* read each signature file */
  MaxFL = -1 ;
  for(i=0; i<(GMM_par->MaxLevels); i++) {
  
      sprintf(parfn, "%s.gmm.s%d.par", GMM_fn,i) ;
      parfd = fopen(parfn, "r");
      if (parfd == NULL) {
          (GMM_par->existSig)[i] = 0 ;   /* no parameters at this level */
          
      } else {
          (GMM_par->existSig)[i] = 1 ;
          S_Pt = &((GMM_par->_Sig)[i]) ;
          I_ReadSigSet(parfd, S_Pt);
          fclose(parfd);

          if((S_Pt->nbands != GMM_par->dimen) ||
             (S_Pt->nclasses != GMM_par->nclasses)) {
              fprintf(stderr,"ERROR: Inconsistent dimen or nclasses in %s\n",
                     parfn);
              exit(10);
          }

          /* test R, calculate log const of Gaussian Density */
          /* containing in Model.c */
          if ( ! extract_init(S_Pt) ) {
              (GMM_par->existSig)[i] = 0 ; 
          } else {
              if(i > MaxFL) MaxFL = i ;
          }
          /*fclose(parfd) ;*/ /* this was alread closed GF 3/25/02 */
      }
  } 
  fclose(fd) ;

  GMM_par->MaxLevels = MaxFL ;

  if(vlevel>0) {
      printf("\tGMM_Par :\n") ;
      printf("\t\tNAME %s\n",  GMM_par->feature_name) ;
      printf("\t\tnbands=%d\n", GMM_par->dimen) ;
      printf("\t\tnClasses=%d\n", GMM_par->nclasses) ;
      printf("\t\tMaxLevels=%d\n", GMM_par->MaxLevels) ;
  } 
}



/*************************************************************************/
/*    write out GMM_par*/
/*************************************************************************/
void WriteGMM_Par( struct Par_GMM *GMM_par, char *GMM_fn ) 
{
  FILE *fd, *parfd ;
  char fn[128], parfn[80] ;
  int i ;

  if(GMM_fn != NULL) {
      sprintf(fn,"%s.gmm", GMM_fn) ;
      if(vlevel>3) printf("\tsave Gaussian mixture model %s\n",fn) ;

      fd = fopen(fn,"w") ;
      if(fd==NULL) FileOpenErr(fn) ;
      
  } else {
      fd= stdout ;
  }
  
  fprintf(fd, "NAME %s\n",  GMM_par->feature_name) ;
  fprintf(fd, "nbands=%d\n", GMM_par->dimen) ;
  fprintf(fd, "nClasses=%d\n", GMM_par->nclasses) ;
  fprintf(fd, "MaxLevels=%d\n", GMM_par->MaxLevels) ;
    
  /* save each signature file */
  for(i=0; i<(GMM_par->MaxLevels); i++) {
      if( (GMM_par->existSig)[i] == 1 ) {
         sprintf(parfn, "%s.gmm.s%d.par", GMM_fn,i) ;
         
         parfd = fopen(parfn, "w");
         if (parfd == NULL) FileOpenErr(parfn);
         I_WriteSigSet(parfd, &((GMM_par->_Sig)[i]));
         fclose(parfd);
      }   
  } 
  
  if(GMM_fn != NULL) fclose(fd) ;
}




/*****************************************************************************/
/*    free quadtree parameters */
/*****************************************************************************/
void FreeGMM_Par( struct Par_GMM *GMM_par ) 
{
  int i ;

  if(vlevel>3) printf("\tfree Gaussian mixture model parameters\n") ;

  /* save each signature file */
  for(i=0; i<(GMM_par->MaxLevels); i++) {
      if( (GMM_par->existSig)[i] == 1 ) {
/*         I_TermSigSet( &((GMM_par->_Sig)[i]) ) ;  */
      } 
  } 
  
  
}



/*****************************************************************************/
/*****************************************************************************/
int write_segmentation(unsigned char **seg, int height, int width, char *fn)
{
  static eisl_img img ;
  static int FirstFlag=1 ;
  static unsigned short bitspersample = 8 ;
  static int KK ;
  int i ;
  
  if(FirstFlag) {   /* generate color map when first using this function */
      img.bitspersample = 8 ;
      KK = (1<<(img.bitspersample)) ;    /* 2^bitspersample */
      img.rmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      img.gmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      img.bmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      if( (img.rmap == NULL) || (img.gmap == NULL) ||
          (img.bmap == NULL) ) {
        fprintf(stderr, "ERROR: Can not allocate color map, write_segment\n");
        return(FALSE) ;
      }
      
      /* Has to be initialied.  o.w. may cause confusion between */
      /* 8-bit and 16-bit color map,  randomly initialization */
      for(i=0; i<KK; i++) {  
          img.rmap[i] = rand() % KK ;
          img.gmap[i] = rand() % KK ;
          img.bmap[i] = rand() % KK ;
      }
      
      /* Since the number of classes are usually less than 10 */
      /* manually set the first 8 colors */
      img.rmap[0]=255;  img.gmap[0]=  0;  img.bmap[0]=  0;  
      img.rmap[1]=  0;  img.gmap[1]=255;  img.bmap[1]=  0;  
      img.rmap[2]=  0;  img.gmap[2]=  0;  img.bmap[2]=255;  
      img.rmap[3]=255;  img.gmap[3]=255;  img.bmap[3]=255;  
      img.rmap[4]=255;  img.gmap[4]=  0;  img.bmap[4]=255;  
      img.rmap[5]=  0;  img.gmap[5]=255;  img.bmap[5]=255;  
      img.rmap[6]=255;  img.gmap[6]=255;  img.bmap[6]=  0;  
      img.rmap[7]=  0;  img.gmap[7]=  0;  img.bmap[7]=  0;  
  }
  
  
  img.height = height ;
  img.width = width ;
  img.img_type = 'p' ;     /* palette image */
  img.img_format = 't' ;   /* tiff image */
  img.mono = seg ;
  
  if( ! write_img (fn, &img, 't', "none") ) { 
       fprintf(stderr,"ERROR: Cannot write out segment %s.\n",fn);
       return(FALSE) ;
  } 
  
  return(TRUE) ;           

}




/*****************************************************************************/
/*****************************************************************************/
void OutputSeg(struct BigPyramid *Pyramid,char *fn, char *String)
{

  int D, nRow,nCol;
  int i, status ;
  unsigned char ***SegPt ;
  char fname[80] ;
  
  if(vlevel>3) printf("\toutput segmentation ......\n") ;

  D = Pyramid->level ;
  SegPt = Pyramid->_Segment ;
  
  /* save the segmentation */    
  for(i=0; i<D; i++) {
    if( (i==0) || ((vlevel >=15)&&(i<3)) || (vlevel > 30)) {    
        nRow = (Pyramid->nRow)[i] ;
        nCol = (Pyramid->nCol)[i] ;
        
        if((SegPt[i])==NULL) {
/*            fprintf(stderr, "WARNING:  NULL pt for segment at level %d, ", i) ;
            fprintf(stderr, "WARNING:  no output at this level !\n" ) ;
*/            
        } else {
    
        	sprintf(fname,"%s.%s.%d",fn,String,i) ;
            if( ! write_segmentation(SegPt[i], nRow, nCol, fname) ) { 
            	fprintf(stderr,"ERROR: Cannot write %s, exit.\n",fname);
            	exit (-10);
        	}            
        	if(vlevel >=3) fprintf(stderr, "\t\t%s is written out\n",fname) ;
       }
    }
  }

}    
       


/*****************************************************************************/
/*****************************************************************************/
void ReadDataWeight(struct ModelPar *Par, char *fn) 
{
    FILE *fd;
    char parfn[80];
    double tmp, *Weight ;
    int i,j,nlevels ;
   
    nlevels = (Par->QTr_par).levels ; 
    Weight=(Par->DataWeight)=(double*) get_spc(nlevels,sizeof(double)) ;
        
    /* read signatures */
    sprintf(parfn, "%s.weight", fn) ;
    if(vlevel >1)  printf("\tReading DataWeights from %s \n", parfn) ;
    fd = fopen (parfn, "r");
    
    if (fd != NULL) {

        /* not include the coarsest level */
        for(i=nlevels-2; i>=0; i--) {    
            fscanf(fd,"level=%d", &j) ;
            if(j!=i) {
              fprintf(stderr, "Wrong format in ReadDataWeight, i%d != j%d\n"
                  ,i,j);
              exit(-1) ;
            }

            fscanf(fd,"%lf\n", &tmp) ;
            Weight[i]=tmp ;

            if(vlevel>20) printf("\t\tDataWeight[%d]=%lf\n",i,tmp) ;
        }
        fclose(fd) ;
        
    } else { 
        
        /* initialize *.weight */
        fprintf(stderr, "WARNING: %s is not found. Set default values\n", 
            parfn) ;
        fd = fopen (parfn, "w");
        if (fd == NULL) FileOpenErr(parfn) ;
        
        tmp = .5 ;
        for(i=(Par->QTr_par.levels-2); i>=0; i--) {    
            fprintf(fd,"level=%d  ", i) ;
            tmp *= .9 ;
            fprintf(fd,"%lf\n", tmp) ;
            Weight[i]=tmp ;

            if(vlevel>50) printf("\t\tDataWeight[%d]=%lf\n",i,tmp) ;
        }
        fclose(fd) ;
    }
}




/*****************************************************************************/
/*****************************************************************************/
void WriteDataWeight(struct ModelPar *Par,  char *fn) 
{
    FILE *fd;
    char parfn[80];
    double *Weight ;
    int i ;
    
    Weight=Par->DataWeight ;
        
    /* read signatures */
    sprintf(parfn, "%s.weight", fn) ;
    if(vlevel >1)  printf("\tWrite out DataWeights into %s\n", parfn) ;
    fd = fopen (parfn, "w");
    if (fd == NULL) FileOpenErr(parfn);

    /* not include the coarsest level */
    for(i=(Par->QTr_par.levels-2); i>=0; i--) 
        fprintf(fd,"level=%d  %lf\n",i,Weight[i]) ;
    fclose(fd) ;
}




/*****************************************************************************/
/* Output Prob Y|X and likelihood */ 
/*****************************************************************************/
void OutputProbLLH(
  struct ModelPar *Par,
  struct BigPyramid *Pyramid,
  char *fn, 
  char *String,
  int x1, int x2,
  int y1, int y2
)
{

  int D, nClasses, nRow,nCol;
  int i,j, m,i1,j1,i2,j2,k ;
  char fname[80] ;
  FILE *outfd ;
  int *ExistSig ;    /* 1 = a signature exists at a level */ 
  double ****LgProb ;     /* log probability of Xn given Yn
                   _LProb[level][j][k][class] */    
  double ****LgLikeH ;     /* LLikeH[level][j][k][class], log likelihood*/
  
  
  if(vlevel>3) printf("\toutput probability ......\n") ;

  D = Pyramid->level ;
  nClasses = (Par->QTr_par).nclasses ;
  LgProb = Pyramid->LProb ;
  LgLikeH = Pyramid->LLikeH ;
  ExistSig = (Par->GMM_par).existSig ;
  
  /* save the data */
      
  i1=x1;  i2=x2; 
  j1=y1;  j2=y2;
  if((i1>i2)||(j1>j2)) {
      fprintf(stderr,"x1 must less that x2 and y1 less than y2\n");
      exit(-10) ;
  }
  
  for(k=0; k<D; k++) {
    nRow = (Pyramid->nRow)[k] ;
    nCol = (Pyramid->nCol)[k] ;
    
    if(i2>(nRow-1)) i2 =nRow-1 ;
    if(j2>(nCol-1)) j2 =nCol-1 ;
    
    sprintf(fname,"%s.%s.Prob%d",fn,String,k) ;
    outfd = fopen(fname, "w");
    if (outfd == NULL) {
             fprintf(stderr,"Can not creat file %s\n",fname);
          exit(10);
    }
      
    for(i=i1; i<=i2; i++)
    for(j=j1; j<=j2; j++) {
                  
        if(ExistSig[k]==1) {
            fprintf(outfd, "%3d  %3d,", i,j) ;
            for(m=0; m<nClasses; m++)
                fprintf(outfd, "(%1d)P=%lf  ",m,LgProb[k][i][j][m]) ;
            fprintf(outfd,"\n") ;
        }

        fprintf(outfd, "%3d  %3d,", i,j) ;
        for(m=0; m<nClasses; m++)
             fprintf(outfd, "(%1d)L=%lf  ",m,LgLikeH[k][i][j][m]) ;
        fprintf(outfd,"\n") ;

    }
      
    fprintf(stderr, "i1=%d, i2=%d, j1=%d, j2=%d, R%d, C%d\n", i1,i2, j1,j2,
        nRow,nCol) ;
    i1 = i1/2 ;  if(i1<0) i1 = 0;
    i2 = i2/2 ;  
    j1 = j1/2 ;  if(j1<0) j1 = 0;
    j2 = j2/2 ;  
     
    fclose(outfd) ;
  }
}    

