#ifndef _TIFF_IMG_
#define _TIFF_IMG_

/*
 * TIFF I/O Library Definitions.
 */
#include "tiffio.h"

#define	TRUE	1
#define	FALSE	0

typedef struct {
  uint32			height;
  uint32			width;
  int 			bitspersample ;
  char			TIFF_type;  	
  					/* 'g' = grayscale;               */
					/* 'b' = bilevel image;           */
					/* 'p' = palette-color;           */
					/* 'c' = RGB full color           */
					/* 'L' = CIEL*a*b* full color     */
					/* 'l' = CIEL*a*b* monochrome     */
					/*	 L* only		  */
					/* 'y' = YCbCr			  */

  unsigned char		**mono;		/* monochrome data, or indices    */
					/* of color-map; indexed as       */
					/* mono[row][col]                 */
  uint16		*rmap, *gmap;   /* RGB color map                  */
  uint16		*bmap;

  uint8			***color;	/* full-color RGB data; indexed   */
					/* as color[plan][row][col],      */
					/* with planes being red,  green, */
					/* and blue, respectively         */
	
  uint8			**luminance;    /* Y component of YCbCr image     */
  					/* L component of La*b* image     */
  int8			***chrominance;	/* chrominance[0][row][col]	  */ 
					/*	Cb of YCbCr image	  */
					/*	a* of La*b* image	  */
					/* chrominance[1][row][col]	  */ 
					/*	Cr of YCbCr image	  */
					/*	b* of La*b* image	  */
  /*				          				  */
  /*  			IMPORTANT NOTE !!!!!				  */
  /*  (1) mono, color[0], and luminance are the SAME pointer		  */
  /*  (2) color[1], and chrominance[0] are the SAME pointer		  */
  /*  (3) color[2], and chrominance[1] are the SAME pointer		  */
  /*  (4) for monochrome images, color, chrominance are not allocated	  */
  /*  (5) luminance is not scaled to its range.  			  */
  /*	  e.g. L of La*b* is usually stored as 0-255 with a unit 0.394531 */
  /*	  when using La*b*, one should scale L back to its range [0,100]  */


  char 			compress_type[256] ;
  /*									  */
  /*	lzw[:opts]	compress output with Lempel-Ziv & Welch encoding  */
  /*	zip[:opts]	compress output with deflate encoding		  */
  /*	jpeg[:opts]compress output with JPEG encoding			  */
  /*	packbits	compress output with packbits encoding		  */
  /*	g3[:opts]	compress output with CCITT Group 3 encoding	  */
  /*	g4		compress output with CCITT Group 4 encoding	  */
  /*	none	use no compression algorithm on output			  */
  /*									  */
  /*	Group 3 options:						  */
  /*	 	1d	use default CCITT Group 3 1D-encoding		  */
  /*		2d	use optional CCITT Group 3 2D-encoding		  */
  /*		fill	byte-align EOL codes				  */
  /*		For example, "g3:2d:fill" to get G3-2D-encoded 		  */
  /*			data with byte-aligned EOLs			  */
  /*									  */
  /*	JPEG options:							  */
  /*		#	set compression quality level (0-100, default 75) */
  /*		r	output color image as RGB rather than YCbCr	  */
  /*		string  file which contains the application supported     */
  /*			quantization tables.				  */
  /*		For example, -c jpeg:r:50:MyQTables  to get JPEG-encoded  */
  /*			RGB data with 50% comp. quality, using q_tables   */
  /*			stored in MyQTables				  */
  /*		NB: The order of parameters are fixed as listed in the    */
  /*			example.					  */
  /*									  */
  /*	LZW and deflate options:					  */
  /*		#	set predictor value				  */
  /*		For example, -c lzw:2 to get LZW-encoded data with 	  */
  /*			horizontal differencing 			  */ 


  /* optinal parameters,  default values are set 			  */
  uint16	photometric ;			/* corresponds to TIFF_type */
  uint16	planarConfig ;
  uint16	samplesperpixel;		/* no default 		  */
  uint8		IsTile;
  uint32	rowsperstrip ;
  uint32	tilewidth ;
  uint32	tildheight ;
  float		XResolution ;			/* number of pixels per   */
						/* ResolutionUnit         */
  float		YResolution ;			/* number of pixels per   */
						/* ResolutionUnit         */
  uint32	ResolutionUnit ;		/* absolute unit of meas. */
  uint16	fillorder ;  			/* 			  */
  /*	    	FILLORDER_MSB2LSB  	1  */	/* most significant -> least */
  /*	    	FILLORDER_LSB2MSB  	2  */	/* least significant -> most */

  /* parameters for JPEG */
  uint8		jpeg_tablesmode ;		/* what to output?       */
  /*		quant_tables		0x0001				 */
  /*		huff_tables		0x0002				 */
  /*		Default is both quant_tables | huff_tables  		 */       
  char*		jpeg_quant_tables_fname ;	/* application supplied  */
						/* quantization tables   */
  uint16**	jpeg_quant_tables ;
  uint16	num_jpeg_quant_tables ;
  uint32	jpeg_quant_tables_length ;

  /*	void*		jpeg_huff_tables ;				 */
  /*	uint32 	jpeg_huff_tables_length ;  				 */

  uint16	horizSubSampling ;		/* for YCbCr image only, */
  uint16	vertSubSampling ;		/* currently, must be 1, */
  						/* unless JPEG compressed */
  
} TIFF_img ;


/* For the following routines: 1 = error reading file; 0 = success  */

/* This routine allocates space and reads TIFF image */
TIFF_img* Read_TIFF_Img(char *fname, int show) ;
int Read_TIFF_Img_Header(TIFF* in, TIFF_img *tiff_img) ;

/* This routine writes out a valid TIFF image */
int Write_TIFF_Img(char *fname, TIFF_img *tiff_img, char *org) ;

/* This routine will allocate one TIFF_img and initial parameters */
/* But it will not allocate any memory for components of TIFF_img */
TIFF_img* New_TIFF_Img(char* name) ;

/* This routine allocates a TIFF image.     */
/* height, width, TIFF_type, bitspersample  */
/* must be defined */
int Allocate_New_TIFF(TIFF_img *tiff_img);

/* This routine frees memory allocated for TIFF image */
void Free_TIFF ( TIFF_img *img );

void *tiff_get_spc(int num, size_t size) ;
void **tiff_get_img(int wd,int ht,size_t size) ;
void tiff_free_img(void **pt) ;
void PrintTIFF_ImgInfo(TIFF_img * tiff_img) ;

#endif /* _TIFF_IMG_ */
