
#include <stdlib.h>
#include <stdio.h>
#include <varargs.h>
#include <math.h>
#include <string.h>
#include "invert.h"

/* Constants */
#define SMPTE_GAMMA 2.2

#define D65X 0.9505
#define D65Y 1.0001
#define D65Z 1.089

#define LUTSIZE 1024

/* Convert Kodak standard gamma corrected RGB to linear RGB.   */
/* Assumes that each input is an array with `length' elements. */
void Kodak_gamma_2_lin(
  unsigned char *rs,unsigned char *gs,unsigned char *bs,
  float *rl, float *gl, float *bl,
  int  length)
{
  int    i;
  double tmp;
  static int    first=1;
  static double lut[256];

  /* precompute constant on first call */
  if(first) {
    for(i=0; i<256; i++) {
      tmp = i*(1/255.0);
      if(tmp<=0.0912) lut[i] = tmp/4;
      else lut[i] = pow( (tmp+0.1115)*(1/1.1115),(1/0.45));
    }
    first = 0;
  }

  for(i=0; i<length; i++) {
    rl[i] = lut[rs[i]];
    gl[i] = lut[gs[i]];
    bl[i] = lut[bs[i]];
  }
}


/* Convert linear RGB to Kodak standard gamma corrected RGB    */
/* Assumes that each input is an array with `length' elements, */
/* and that each element is >0 and <1 .                        */
int lin_2_Kodak_gamma(
  float *rl, float *gl, float *bl,
  unsigned char *rs, unsigned char *gs, unsigned char *bs,
  int  length)
{
  int    i;
  int    index;
  double tmp;
  int    limit_flag;
  static int    first=1;
  static unsigned char lut[LUTSIZE];

  /* precompute constant on first call */
  if(first) {
    for(i=0; i<LUTSIZE; i++) {
      tmp = (i + 0.5)/LUTSIZE;
      if(tmp<=0.0228) lut[i] = 255*(4.0* tmp) + 0.5;
      else lut[i] = 255*(1.1115*pow(tmp,0.45) - 0.1115) + 0.5;
    }
    first = 0;
  }

  /* Apply LUT; set limit flag if overflow occurs */
  limit_flag=0;
  for(i=0; i<length; i++) {
    index = LUTSIZE*rl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag=1; }
    else
      if(index<0) { index = 0; limit_flag=1; }
    rs[i] = lut[index];

    index = LUTSIZE*gl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag=1; }
    else
      if(index<0) { index = 0; limit_flag=1; }
    gs[i] = lut[index];

    index = LUTSIZE*bl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag=1; }
    else
      if(index<0) { index = 0; limit_flag=1; }
    bs[i] = lut[index];
  }

  return(limit_flag);
}


/* Convert gamma corrected RGB to linear RGB.   */
/* Assumes that each input is an array with `length' elements. */
void gamma_2_lin(
  unsigned char *rs,unsigned char *gs,unsigned char *bs,
  float *rl, float *gl, float *bl,
  double gamma,
  int  length)
{
  int    i;
  static int    first=1;
  static double lut[256];

  /* precompute constant on first call */
  if(first) {
    for(i=0; i<256; i++) lut[i] = pow( i*(1/255.0),gamma);
    first = 0;
  }

  for(i=0; i<length; i++) {
    rl[i] = lut[rs[i]];
    gl[i] = lut[gs[i]];
    bl[i] = lut[bs[i]];
  }
}


/* Gamma and offset correct linear RGB components.                  */
/* This program is used to calibrate data for a particular monitor. */
/* Assumes that each input is an array with `length' elements, */
/* and that each element is >0 and <1 .                        */
int lin_2_gamma(
  float *rl, float *gl, float *bl,
  unsigned char *rs, unsigned char *gs, unsigned char *bs,
  double gamma, double offset,
  int  length)
{
  int    i;
  int    index;
  double tmp;
  int    limit_flag;
  static int    first=1;
  static unsigned char lut[LUTSIZE];

  /* precompute constant on first call */
  if(first) {
    for(i=0; i<LUTSIZE; i++) {
      tmp = (i + 0.5)/LUTSIZE;
      lut[i] = (255.0-offset)*pow(tmp,1.0/gamma) + offset + 0.5;
    }
    first = 0;
  }

  /* Apply LUT; set limit flag if overflow occurs */
  limit_flag = 0;
  for(i=0; i<length; i++) {
    index = LUTSIZE*rl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag = 1; }
    else
      if(index<0) { index = 0; limit_flag = 1; }
    rs[i] = lut[index];

    index = LUTSIZE*gl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag = 1; }
    else
      if(index<0) { index = 0; limit_flag = 1; }
    gs[i] = lut[index];

    index = LUTSIZE*bl[i];
    if(index>=LUTSIZE) { index = LUTSIZE-1; limit_flag = 1; }
    else
      if(index<0) { index = 0; limit_flag = 1; }
    bs[i] = lut[index];
  }

  return(limit_flag);
}



/* Transforms [a1,a2,a3] to [b1,b2,b3] using the transformation matrix A. */
/* Assumes that each input is an array with `length' elements. */
void color_transform(
  float *a1,float *a2,float *a3,
  float *b1,float *b2,float *b3,
  double A[3][3],
  int length)
{
  int i;
  double c1,c2,c3;

  for(i=0; i<length; i++) {
    c1 = A[0][0]*a1[i] + A[0][1]*a2[i] + A[0][2]*a3[i];
    c2 = A[1][0]*a1[i] + A[1][1]*a2[i] + A[1][2]*a3[i];
    c3 = A[2][0]*a1[i] + A[2][1]*a2[i] + A[2][2]*a3[i];

    b1[i] = c1;
    b2[i] = c2;
    b3[i] = c3;
  } 
}



/* Returns the matrix which transforms from linear SMPTE RGB             */
/* with D65 white point to XYZ. The input and output are column vectors. */ 
void init_RGB_2_XYZ(double A[3][3])
{
  A[0][0]=0.3935; A[0][1]=0.3653; A[0][2]=0.1917; 
  A[1][0]=0.2124; A[1][1]=0.7011; A[1][2]=0.0866; 
  A[2][0]=0.0187; A[2][1]=0.1119; A[2][2]=0.9584; 
}


/* Returns the matrix which transforms from Mullen's opponent color space */
/* (defined in Bernd's paper) to XYZ.                                     */
void init_Yrgyb_2_XYZ(double A[3][3])
{
  A[0][0]=0.0670; A[0][1]=-0.1116; A[0][2]=-0.0830; 
  A[1][0]=0.0707; A[1][1]=0.0; A[1][2]=0.0; 
  A[2][0]=0.0767; A[2][1]=0.0049; A[2][2]=-1.0; 
}


void init_XYZ_2_Ycxcz(double A[3][3])
{
  A[0][0]=(1/D65X)*0.0; A[0][1]=(1/D65Y)*1.16; A[0][2]=(1/D65Z)*0.0; 
  A[1][0]=(1/D65X)*5.0; A[1][1]=(1/D65Y)*-5.0; A[1][2]=(1/D65Z)*0.0; 
  A[2][0]=(1/D65X)*0.0; A[2][1]=(1/D65Y)*2.0;  A[2][2]=(1/D65Z)*-2.0; 
}



/* multiplies color transformations to form A = BC */
void maxtrix_matrix_mult(double A[3][3],double B[3][3],double C[3][3])
{
  int i,j,k;

  for(i=0; i<3; i++)
  for(j=0; j<3; j++) {
    A[i][j] = 0;
    for(k=0; k<3; k++) A[i][j] += B[i][k]*C[k][j];
  }
}


void invert_xform(double A[3][3])
{
  invert3((double *)A,3);
}


/* Uses D65 white point to convert XYZ colors to Lab.          */
/* Assumes that each input is an array with `length' elements. */
void XYZ_2_Lab(
  float *X,float *Y,float *Z,
  float *L,float *a,float *b,
  int length)
{
  int           i;
  static int    first=1;
  static double Yn_inv,Xn_inv,Zn_inv;
  double        ty,tx,tz;

  /* precompute constant on first call */
  if(first) {
    /* D65 White point correction */
    Xn_inv = 1.0/D65X;
    Yn_inv = 1.0/D65Y;
    Zn_inv = 1.0/D65Z;
    first = 0;
  }

  for(i=0; i<length; i++) {
    ty = Y[i]*Yn_inv;
    if(ty<0.008856) ty = 7.787*ty + (16.0/116.0);
    else ty = pow(ty,1.0/3.0);

    tx = X[i]*Xn_inv;
    if(tx<0.008856) tx = 7.787*tx + (16.0/116.0);
    else tx = pow(tx,1.0/3.0);

    tz = Z[i]*Zn_inv;
    if(tz<0.008856) tz = 7.787*tz + (16.0/116.0);
    else tz = pow(tz,1.0/3.0);

    L[i]=116.0*ty - 16;
    a[i]=500.0*(tx - ty);
    b[i]=200.0*(ty - tz);
  }
}



/* Uses D65 white point to convert Lab colors to XYZ */
/* Assumes that each input is an array with `length' elements. */
void Lab_2_XYZ(
  float *L,float *a,float *b,
  float *X,float *Y,float *Z,
  int length)
{
  int     i;
  static  int first=1;
  static  double f_boundary;
  double  ty,tx,tz;

/* Precompute comparison constant on first call */
  if(first) {
    f_boundary = 7.787 * (0.008856) + (16.0/116.0);
    first = 0;
   }
    
  for(i=0; i<length; i++) {
    ty = (L[i] + 16.0)/116.0;
    tx = a[i]/500 + ty;
    tz = -b[i]/200 + ty;

    if(ty < f_boundary) ty = (ty - (16.0/116.0))/7.787;
    else ty = pow(ty,3.0);

    if(tx < f_boundary) tx = (tx - (16.0/116.0))/7.787;
    else tx = pow(tx,3.0);

    if(tz < f_boundary) tz = (tz - (16.0/116.0))/7.787;
    else tz = pow(tz,3.0);

    Y[i] = ty * D65Y;
    X[i] = tx * D65X;
    Z[i] = tz * D65Z;
  }
}

